from Definitions import LorRepDefinitions, EnergyDimensions
from readers import Reader
from utility import convertfrac

import sympy

class Container:
    """
    Parent class for all the different containers.
    """


    #init statement
    def __init__(self):
        pass


class FieldsContainer(Container):

    """
    This class is the core piece of the 3 step procedure from Feynrules model file to finished effective Lagrangian. It receives
    the information read by the Reader class and transforms it in a way that makes it serve as DEFT input.

    The Reader instance is providing all the replacement rules in the form of a dictionary. E.g.

    Replacement Rule:


                Classname -> Phi,
                SelfConjugate -> True

    Reader Output:

                {'Classname': 'Phi','SelfConjugate':'True'}

    For each of the possible Replacement rules, there is a '...Converter' function, that takes care of the specific 
    replacement rule and transforms it into a form that is resembling the DEFT input parameters. There are dummy 
    '...Converter' functions for all of the parameters that are not used by DEFT, such that in principle, a fully fledged
    Feynrules model file can be used.

    Additionally, there are two 'Build...' funcntions that provide the dictionary specifying the fields transformation properties 
    under non abelian gauge groups as well as the associated field strength of a gauge boson.

    """



    #here are all the converter functions that are being called by init to transform the model file inputs into DEFT specific information

    def QuantumNumbersConverter(self,quantumnumbers):
        NewReader = Reader()
        U1dict = NewReader.ReplacementToDictionary(quantumnumbers)
        for key in U1dict:
            U1dict[key] = convertfrac(U1dict[key])
        self.uones = U1dict

    def ClassNameConverter(self,classname):
        self.classname = classname
        self.name = classname
        self.basename = classname
    
    def TeXParticleNameConverter(self,TeX):
        #remove the unnecessary [] brackets
        texstart = TeX.find('\[')
        while texstart != -1:
            texend = TeX[texstart:].find(']')
            realtexend = texend + texstart
            TeX = TeX[:texstart+1] + TeX[texstart+2:realtexend] + TeX[realtexend+ 1:]
            texstart = TeX.find('\[')
        
        #check if is subscript
        substart = TeX.find('Subscript[')
        if substart != -1:
            subend = substart + len('Subscript[')
            comma = TeX[subend:].find(',')
            realcomma = comma + subend
            closingbracket = TeX[realcomma:].find(']')
            realclosingbracket = closingbracket + realcomma

            name = TeX[subend:realcomma]
            subscript = TeX[realcomma + 1:realclosingbracket]
            TeX = name +"_{"+subscript+"}"

        #check if is superscript
        supstart = TeX.find('Superscript[')
        if supstart != -1:
            supend = supstart + len('Superscript[')
            comma = TeX[supend:].find(',')
            realcomma = comma + supend
            closingbracket = TeX[realcomma:].find(']')
            realclosingbracket = closingbracket + realcomma

            name = TeX[supend:realcomma]
            supscript = TeX[realcomma + 1:realclosingbracket]
            TeX = name +"^{"+supscript+"}"

        self.texname = TeX

    def SelfConjugateConverter(self,selfconjugate):
        if selfconjugate == "True":
            self.isSelfConjugate = True
        elif selfconjugate == "False":
            self.selfconjugate = False
        else:
            raise Exception('The value of Selfconjugate has to be either True or False!')

    def FieldTypeConverter(self,fieldtype):
        #determine the lorentz representation
        if fieldtype == 'W' and self.definitiondict['Chirality'] == 'Left':
            fieldtype = 'WL'
        elif fieldtype == 'W' and self.definitiondict['Chirality'] == 'Right':
            fieldtype = 'WR'
        self.nonabel.update(LorRepDefinitions[fieldtype])

        #determine the dimension

        self.dimension = EnergyDimensions[fieldtype]

    def IndicesConverter(self,indices):

        indnames = []

        done = False
        while not done:
            startindex = indices.find('Index[')
            if startindex == -1:
                done = True
            else:
                realstartindex = startindex + len('Index[')
                endindex = indices[realstartindex:].find(']')
                realendindex = endindex + realstartindex
                indname = indices[realstartindex:realendindex]
                indnames.append(indname)
                indices = indices[realendindex:]
        
        self.indices = indnames

    def MassConverter(self,Mass):
        pass
    
    def WidthConverter(self,Width):
        pass

    def ParticleNameConverter(self,ParticleName):
        pass
    
    def PDGConverter(self,PDG):
        pass

    def PropagatorLabelConverter(self,PropagatorLabel):
        pass

    def PropagatorTypeConverter(self,PropagatorType):
        pass
    
    def PropagatorArrowConverter(self,PropagatorArrow):
        pass

    def FullNameConverter(self,FullName):
        pass

    def FlavorIndexConverter(self,FlavorIndex):
        pass
    
    def DefinitionsConverter(self,Definitions):
        pass
    
    def AntiParticleNameConverter(self,AntiParticleName):
        pass

    def ClassMembersConverter(self,ClassMembers):
        pass

    def GoldstoneConverter(self,Goldstone):
        pass

    def UnphysicalConverter(self,unphysical):
        pass
    
    def ChiralityConverter(self,chirality):
        pass


    # these are the builder functions

    def BuildNonAbels(self,groupname,repnames):
        """
        Description:    This method builds the dictionary specifying the representation under a gauge group
        Input:          The name 'groupname' of the group the representation dictionary needs to be build for
                        A dictionary repnames
        

        """
        hasnontrivialrep = False
        for key in repnames:
            if key in self.indices:
                self.nonabel.update({groupname:repnames[key]})
                hasnontrivialrep = True
        
        if not hasnontrivialrep:
            self.nonabel.update({groupname:'1'})

    def BuildFieldStrength(self,groupname,isabelian):
        """
        This method generates all the informations needed for the fielddeclaration class to generate a FieldStrength entry in the fields file.
        The information is returned in the form of a dictionary.
        """
        FSdictionary = {}

        classname = self.classname[1:]
        dimension = '2'

        #build the nonabel dictionary for the field strengths
        nonabel = {}
        nonabel.update(self.nonabel)
        if not isabelian:
            nonabel[groupname] = 'FS'
        nonabel['lorL'] = '3'
        nonabel['lorR'] = '1'

        uones = {}
        uones.update(self.uones)
        name = classname
        basename = classname
        texname = classname
        selfconjugate = True

        if isabelian:
            UoneFS = True
        else:
            UoneFS = False

        vectorpotential = self.classname 
        
        FSdictionary = {'cn':classname,'dim':dimension,'nonabel':nonabel,'uones':uones,'n':name,'bn':basename,'tn':texname,'sc':selfconjugate,'UoneFS':UoneFS,'vp':vectorpotential}

        return FSdictionary

    #init statement

    def __init__(self,definitiondict):

        #these are the quantities that are read from the model file
        
        self.definitiondict = definitiondict

        #these are the quantities that have to be transformed and will be used by the writetoFieldsFile class methods
        #the transformation is done by internal class methods

        self.classname = ""
        self.dimension = 0
        self.nonabel = {}
        self.uones = {}
        self.name = ""
        self.basename = ""
        self.texname = ""
        self.isSelfConjugate = False
        #self.isU1FieldStrength = False
        #self.vectorpotential = None
        
        self.indices = []

        #call a converter function for each of the dictionary items that are in definitiondict
        for key in self.definitiondict:
            eval("self." + key+"Converter("+"\""+self.definitiondict[key]+"\")")


class GaugeGroupsContainer(Container):

    #here are all the converter functions that are being called by init to transform the model file inputs into DEFT specific information

    def AbelianConverter(self,Abelian):
        if Abelian == "True":
            self.isAbelian = True
        elif Abelian == "False":
            self.isAbelian = False
        else:
            raise Exception("Abelian has to be either True or False!")

    def GaugeBosonConverter(self,GaugeBoson):
        if self.Groupdefinitiondict['Abelian'] == 'True':
            key = self.Groupdefinitiondict['Charge']
            newdictionary = {key:"fields."+GaugeBoson[1:]} #---> together with the naming convention, this ensure that the field strength name is selected here
            self.GaugeGroupField.update(newdictionary)     #     e.g. gaugeboson: AB (this is the vector), the field strength is then B
        elif self.Groupdefinitiondict['Abelian'] == 'False':
            key = self.Groupdefinitiondict['GroupName']
            newdictionary = {key:"fields."+GaugeBoson[1:]} #---> together with the naming convention, this ensure that the field strength name is selected here
            self.GaugeGroupField.update(newdictionary)
        else:
            raise Exception('Abelian has to be either true or false!')
        
        self.gaugeboson = GaugeBoson

    def ChargeConverter(self,Charge):
        self.Charge = Charge

    def CouplingConstantConverter(self,Couplingconstant):
        if self.Groupdefinitiondict['Abelian'] == 'True':
            key = self.Groupdefinitiondict['Charge']
            newdictionary = {key:Couplingconstant}
            self.U1GaugeGroupCoupling.update(newdictionary)
        elif self.Groupdefinitiondict['Abelian'] == 'False':
            key = self.Groupdefinitiondict['GroupName']
            newdictionary = {key:Couplingconstant}
            self.NAGaugeGroupCoupling.update(newdictionary)
        else:
            raise Exception('Abelian has to be either true or false!')

    def GroupNameConverter(self,GroupName):
        if self.Groupdefinitiondict['Abelian'] == 'True':
            self.groupname = self.Groupdefinitiondict['Charge']
        elif self.Groupdefinitiondict['Abelian'] == 'False':
            self.groupname = GroupName
        else:
            raise Exception('Abelian has to be either true or false!')

    def FundamentalDimensionConverter(self,FundamentalDimension):

        self.dimfund = int(FundamentalDimension)

    def StructureConstantConverter(self,StructureConstant):
        pass

    def RepresentationsConverter(self,Representations):
        representationlist = []
        representationlisttext = Representations[1:-1] #get rid of the outermost braces
        done = False
        while not done:
            comma = representationlisttext.find(',')
            if comma == -1:
                break
            end = representationlisttext.find('}')
            if end == -1:
                break
            generator = representationlisttext[1:comma]
            indexname = representationlisttext[comma+1:end]
            representationlist.append((generator,indexname))
            representationlisttext = representationlisttext[end:]
            separatingcomma = representationlisttext.find(',')
            if separatingcomma == -1:
                break
            else:
                representationlisttext = representationlisttext[separatingcomma+1:]
        
        self.representations.update({self.groupname:representationlist})

    def DefinitionsConverter(self,Definitions):
        pass

    ########## end of converter functions #######################


    #These are the builder methods that take the retreived gauge group information and create the .json files that store the DEFT variables

    def BuildGroupReps(self):
        

        #build the groupreps dictionary
        reps = {}
        for key in self.relevantIndices:
            reps.update({self.relevantIndices[key]['IndexRange']:(1,0,[])})

        #add a groupref for the field strength
        reps.update({'FS':(1,1,[('Traceless',[0,1])])})

        #add a groupdef for the trivial rep

        reps.update({'1':(0,0,[])})

        groupdef = {self.groupname:reps}
        self.groupdef = groupdef

        return groupdef
        
    def BuildGroupsDimFund(self):
        groupsdimfund = {self.groupname:self.dimfund}
        self.groupsdimfund = groupsdimfund
        return groupsdimfund

    def BuildTeXIndexDict(self):
        """
        This method generates the indices that are used for the Tex representation in DEFT.
        """

        TeXIndices = {}
        

        #determine the startindex

        startind = self.Indexdefinitiondicts[self.Indexdefinitiondicts.keys()[0]]['IndexStyle']

        indlist = list(map(chr, range(ord(startind), ord(startind)+11)))
        
        TeXIndices.update({self.groupname:indlist})

        self.TeXindices = TeXIndices
        return TeXIndices

    def BuildRepresentationNames(self):
        """
        make a dictionary that relates the index names to the representation names 
        is of the form {'groupname':('Indexname','repname'),...}
        """

        repnames = {self.groupname:{}}

        for key in self.relevantIndices:
            repnames[self.groupname].update({key:self.relevantIndices[key]['IndexRange']})
        
        self.repnames = repnames
        return repnames

    ########## end of the builder methods ##########################


    # init statement

    def __init__(self,Groupdefinitiondict,Indexdefinitiondicts):


        self.Groupdefinitiondict = Groupdefinitiondict
        self.Indexdefinitiondicts = Indexdefinitiondicts

        self.gaugeboson = ""

        self.dimfund = 1

        self.representations = {}
        

        #these are the quantities that were transformed and will be used to implement the gauge groups into DEFT
        #the transformation is done by internal class methods

        self.groupname = ""
        self.isAbelian = True
        self.U1GaugeGroupCoupling = {}
        self.NAGaugeGroupCoupling = {}
        self.GaugeGroupField = {}
        self.groupdef = {}
        self.repnames = {}
        self.groupsdimfund = {}
        self.TeXindices = {} # this is the dictionary containing the indexlist for the TeX Representation
        self.relevantIndices = {}



        #call a converter function for each of the dictionary items that are in definitiondict

        #for the groups
        for key in self.Groupdefinitiondict:
            eval("self." + key+"Converter("+"\""+self.Groupdefinitiondict[key]+"\")")

        #determine the indices relevant for this group and store them in a class variable

        for key in self.Indexdefinitiondicts:
            if key.find(self.groupname) == 0:
                self.relevantIndices.update({key:self.Indexdefinitiondicts[key]})
        



        #run the builder functions

        if not self.isAbelian:
            self.BuildGroupReps()
            self.BuildRepresentationNames()
            self.BuildGroupsDimFund()
            self.BuildTeXIndexDict()


        

        



