#This file holds the class definitions for the different classes that are used to read out the
#model file and store them into instances of the container classes

class Reader:
    """
    Parent class for all the Reader classes
    This class is the first step in the three step procedure from FeynRules model file to final effective Lagrangian.
    It receives the full text of the FeynRules model file as input and returns the information from the replacement 
    rules in the form of dictionaries: E.g.

    Replacement Rules:


                Classname -> Phi,
                SelfConjugate -> True

    Reader Output:

                {'Classname': 'Phi','SelfConjugate':'True'}
    """

    def ReplacementToDictionary(self,ReplacementString):
        """
        Description:    This method transforms a string like {classname -> phi, charge -> {Y -> 1.0}, components -> {R1,R2}} to a dictionary like
                        {'classname':'phi','charge':'{Y->1.0}','components':'{R1,R2}'}
                        This is supposed to be used at first level and can be applied subsequently to the dictionary items to transform the remaining
                        replacement rules like {Y->1.0} into {'Y':'1.0'}
        Input:          'ReplacementString' containing the replacement rules. Should be of the form: {...->..} and not contain any white spaces or
                        line breaks
        Output:         A dictionary that has transformed the information from the FIRST LAYER of replacement rules into dictionary form, e.g.
                        {'classname':'phi','charge':'{Y->1.0}','components':'{R1,R2}'}
        """

        #strip beginning and ending { }
        ReplacementString = ReplacementString[1:-1]

        #check if there are any replacement rules in the string
        if ReplacementString.find('->') == -1:
            raise Exception('No replacement rules in ReplacementString input!')
        DictKeys = [] #keys for the dictionary that is being returned
        DictItems = [] #items for the dictionary that is being returned
        Dict = {} #dictionary that ultimatively gets returned

        done = False

        while not done:
            ArrowIndex = ReplacementString.find('->')
            DictKey = ReplacementString[:ArrowIndex]
            DictKeys.append(DictKey)
            ReplacementString = ReplacementString[ArrowIndex+2:]
            CommaIndex = ReplacementString.find(',')
            #check, if we are already at the end
            if CommaIndex == -1:
                DictItem = ReplacementString
                DictItems.append(DictItem)
                done = True
            else:
                #make sure the commaIndex is not from a comma inside a new list or dicitonary
                if ReplacementString[:CommaIndex].find('{') == -1 and ReplacementString[:CommaIndex].find('[') == -1:
                    DictItem = ReplacementString[:CommaIndex]
                    DictItems.append(DictItem)
                    ReplacementString = ReplacementString[CommaIndex+1:]
                elif ReplacementString[:CommaIndex].find('{') == -1 and ReplacementString[:CommaIndex].find('[') != -1:
                    numofopeningbrackets = 1
                    numofclosingbrackets = 0

                    FirstParIndex = ReplacementString.find('[')

                    counter = FirstParIndex + 1
                    while numofclosingbrackets != numofopeningbrackets:
                        if ReplacementString[counter] == '[':
                            numofopeningbrackets += 1
                        elif ReplacementString[counter] == ']':
                            numofclosingbrackets += 1
                        counter +=1
                    
                    #check again if we are already at the end
                    if ReplacementString[counter:].find(',') == -1:
                        DictItem = ReplacementString
                        DictItems.append(DictItem)
                        done = True
                    else:
                        RealCommaIndex = ReplacementString[counter:].find(',') + counter
                        DictItem = ReplacementString[:RealCommaIndex]
                        DictItems.append(DictItem)
                        ReplacementString = ReplacementString[RealCommaIndex+1:]
                #in case there is a {} and a [] (then the {} will always open before the [] and we can treat it as if there were no [])
                else:
                    numofopeningbrackets = 1
                    numofclosingbrackets = 0

                    FirstParIndex = ReplacementString.find('{')

                    counter = FirstParIndex + 1
                    while numofclosingbrackets != numofopeningbrackets:
                        if ReplacementString[counter] == '{':
                            numofopeningbrackets += 1
                        elif ReplacementString[counter] == '}':
                            numofclosingbrackets += 1
                        counter +=1
                    
                    #check again if we are already at the end
                    if ReplacementString[counter:].find(',') == -1:
                        DictItem = ReplacementString
                        DictItems.append(DictItem)
                        done = True
                    else:
                        RealCommaIndex = ReplacementString[counter:].find(',') + counter
                        DictItem = ReplacementString[:RealCommaIndex]
                        DictItems.append(DictItem)
                        ReplacementString = ReplacementString[RealCommaIndex+1:]
        for i, item in enumerate(DictItems):
            Dict[DictKeys[i]] = item
            
        return Dict

    def RemoveSpaces(self,string):
        """
        Description:    This method will remove all the spaces and newlines from the string.
        Input:          A string 'string' one wants to relief from all the spaces and newlines
        Output:         The contents of 'string' but with no spaces and newlines
        """

        newstring = ""
        
        for i in xrange(len(string)):
            if string[i] != " " and string[i] != "\n":
                newstring += string[i]

        newstring = newstring.rstrip()
        
        return newstring

    def RemoveCommentary(self,string):
        """
        Description:    This method removes all commentary (in the form of (*.......*) ) from the string "string".
        Input:          a 'string' possibly containing commentary
        Output:         the string 'string' where there are no more commentary lines of the form (*.........*)
        """

        done = False

        while not done:
            commentarystart = string.find("(*")
            if commentarystart == -1:
                done = True
            else:
                commentaryend = string.find("*)")
                string = string[:commentarystart] + string[commentaryend+2:]
        
        return string

    def ReadModelFilewoSpaces(self,modelfilepath):
        """
        Description:    This method is used to read in the whole modelfile stored at the modelfilepath. The text of the model file
                        will be stored at the class variable 'self.ModelFileText' WITHOUT any spaces or newlines.
        Input:          The path to the model file 'modelfilepath' one wants to read in
        Output:         -
        """

        with open(modelfilepath, "r") as readfile:
            self.ModelFileText = self.RemoveCommentary(self.RemoveSpaces(readfile.read()))

    def ReadModelFilewSpaces(self,modelfilepath):
        """
        Description:    This method is used to read in the whole modelfile stored at the modelfilepath. The text of the model file
                        will be stored at the class variable 'self.ModelFileText'.
        Input:          The path to the model file 'modelfilepath' one wants to read in
        Output:         -
        """

        with open(modelfilepath, "r") as readfile:
            self.ModelFileText = readfile.read()


    def WriteModelFile(self,modelfilepath):
        """
        Description:    This method is used to write the string in the internal variable 'self.ModelFileText'
                        to the file at the location "modelfilepath"
        Input:          The path to the model file 'modelfilepath' one wants to write to
        Output:         -
        """

        with open(modelfilepath, "w") as writefile:
            writefile.write(self.ModelFileText)
  
    def PrintModelFileText(self):
        """
        Description:    This method prints the string currently stored in the class variable "ModelFileText"
        Input:          -
        Output:         -
        """

        print(self.ModelFileText)
        
    def ExtractDefinitions(self,title):
        """
        Description:    This method extracts the bit of the definitions that belongs to "title" (e.g. the ClassesDescriptions or the GaugeGroups) from
                        the full ModelFileText currently stored in the class variable self.ModelFileText (!should not contain any spaces or new lines anymore!)
        Input:          The title 'title' of the section of definitionsi, i.e. 'M$Parameters','M$GaugeGroups','M$ClassesDescription'.
        Output:         The content of the section of definitions belonging to 'title' without beginning and ending '{' '}' (nor spaces or newlines), e.g.

                        M$Parameters = {
                                        gs1 == { 
                                                    ParameterType       -> Internal, 
                                                    TeX                 -> Subscript[g,s1]
                                                },

                                        gw == { 
                                                    ParameterType       -> Internal, 
                                                    TeX                 -> Subscript[g,w]
                                                }
                                        }

                        becomes:

                        gs1 == {ParameterType->Internal,TeX->Subscript[g,s1]},gw=={ParameterType->Internal,TeX->Subscript[g,w]}

        """

        startindex = self.ModelFileText.find(title)
        DefinitionText = self.ModelFileText[startindex:]

        #determine position of ending }
        
        start = DefinitionText.find("{") + 1 

        searchstring = DefinitionText[start:]

        numofopeningbrackets = 1
        numbofclosingbrackets = 0
        counter = 0

        while numbofclosingbrackets != numofopeningbrackets:
            if searchstring[counter] == "{" or searchstring[counter] == "}":
                if searchstring[counter] == "{":
                    numofopeningbrackets +=1
                if searchstring[counter] == "}":
                    numbofclosingbrackets += 1
            counter += 1
            

        ending = counter - 1
        DefinitionText = searchstring[:ending]
        return DefinitionText

    def ReplaceDefinitions(self,title,newdefinitions):
        """
        Description:    This method allows one to swap the existing definition in the modelfiletext (e.g. M$Parameters) with the newone in newdefinitions.
                        It is used to update the parameters list with the coefficients of the various terms that have been generated.
        Input:          The title 'title' of a section of descriptions, i.e. 'M$Parameters','M$GaugeGroups','M$ClassesDescription'
                        The text of the new definitions 'newdefinitions' that is supposed to replace the old replacement rules in the section labeled by 'title'
        Output:         -
        """

        startindex = self.ModelFileText.find(title)


        DefinitionText = self.ModelFileText[startindex:]

        start = DefinitionText.find("{") + 1 

        searchstring = DefinitionText[start:]

        numofopeningbrackets = 1
        numbofclosingbrackets = 0
        counter = 0

        while numbofclosingbrackets != numofopeningbrackets:
            if searchstring[counter] == "{" or searchstring[counter] == "}":
                if searchstring[counter] == "{":
                    numofopeningbrackets +=1
                if searchstring[counter] == "}":
                    numbofclosingbrackets += 1
            counter += 1
            

        endindex = counter -1 + startindex + start

        NewModelfiletext = self.ModelFileText[:startindex] + newdefinitions + self.ModelFileText[endindex:]

        self.ModelFileText = NewModelfiletext

    def DivideDefinitions(self,definitionstring):

            """
            Description:    This function divides the different field definitions into separate strings and stores 
                            them all in a list that is returned.
            Input:          The string 'definitionstring' containing all the replacement rules belonging to all the fields or all the groups
            Output:         A list 'DefinitionStrings' which contains strings, each of them being the replacement rules of a single field or group as well as their assignment, e.g.
                            [ 'W[6]=={ClassName->DummyL,SelfConjugate->False,Chirality->Left,QuantumNumbers->{Y->0}}',...]
            """

            definitionstring = self.RemoveSpaces(definitionstring)

            DefinitionStrings = []

            done = False

            while not done:
                
                numofopeningbrackets = 1
                numofclosingbrackets = 0
                start = definitionstring.find("{")
                counter = start + 1
                while numofopeningbrackets != numofclosingbrackets:
                    if definitionstring[counter] == "}" or definitionstring[counter] == "{":
                        if definitionstring[counter] == "}":
                            numofclosingbrackets += 1
                        elif definitionstring[counter] == "{":
                            numofopeningbrackets += 1
                    counter += 1
                
                CommaIndex = definitionstring[(counter-1):].find(",")
                fullCommaIndex = CommaIndex + counter -1

                if CommaIndex == -1:
                    DefinitionStrings.append(definitionstring)
                    done = True
                else:
                    DefinitionStrings.append(definitionstring[:fullCommaIndex])
                    definitionstring = definitionstring[fullCommaIndex+1:]
                    

            return DefinitionStrings

    #init statement
    def __init__(self):
        self.ModelFileText = ""

class FieldsReader(Reader):

    def ExtractFieldDefinitions(self,fulldefinitionstring):

        """
        Description:    This method extracts each individual field definitions and stores the result in a list
                        of dictionaries (one dict for each field containing all the information about the field 
                        such as "classname" and "selfconjugate")
        Input:          The string containig all the replacement rules for all of the field of 'M$ClassesDescription'
        Output:         A list of dictionaries 'definitionDictionaries', one for each field that is defined. Each dictionariy contains all the
                        information from the replacement rules for this field (including the ones not relevant for DEFT)        
        """


        def DetermineFieldType(stringbit):
            """
            Description:    This function reads the field type of the field from its definitionstring
            Input:          The definition string of a field beginning with e.g. 'W[6]==...'
            Output:         The letter representing the field type of the field, e.g. W, F, V,...
            """

            if stringbit[1] == "[":
                FieldType = stringbit[0]
            else:
                FieldType = stringbit[:2]
            return FieldType

        #divide the string containing all the field definitions in to a list where each item represents the definition of one field
        fieldDefintions = self.DivideDefinitions(fulldefinitionstring)

        definitionDictionaries = []

        #transform the strings into dictionaries. 
        for defs in fieldDefintions:
            fieldtype = DetermineFieldType(defs)
            startIndex = defs.find("{")
            replacementRules = defs[startIndex:]
            defdictionary = self.ReplacementToDictionary(replacementRules)
            defdictionary.update({'FieldType':fieldtype})
            definitionDictionaries.append(defdictionary)
        
        self.FieldDefinitions.append(definitionDictionaries)


        return definitionDictionaries

    #init statement
    def __init__(self):
        self.FieldDefinitions = []
    
class GaugeGroupsReader(Reader):

    def ExtractGroupDefinitions(self,fulldefinitionstring):
        """
        Description:    This method extracts each individual group definitions and stores the result in a list
                        of dictionaries (one dict for each group containing all the information about the group
                        such as "classname" and "selfconjugate")
        Input:          The string containig all the replacement rules for all of the groups of 'M$GaugeGroups'
        Output:         A list of dictionaries 'definitionDictionaries', one for each group that is defined. Each dictionariy contains all the
                        information from the replacement rules for this group (including the ones not relevant for DEFT)
        """

        def DetermineGroupName(stringbit):
            """
            Description:    This function determines the name of the group.
            Input:          The string 'stringbit' beginning with e.g. 'U1Y == {...'
            Output:         The groupname 'Groupname', e.g. 'U1Y'
            """

            equalindex = stringbit.find('=')
            Groupname = stringbit[:equalindex]
            Groupname = self.RemoveSpaces(Groupname)

            return Groupname

        groupDefintions = self.DivideDefinitions(fulldefinitionstring)

        definitionDictionaries = []

        #transform the strings into dictionaries.
        for defs in groupDefintions:
            groupname = DetermineGroupName(defs)
            startIndex = defs.find("{")
            replacementRules = defs[startIndex:]
            defdictionary = self.ReplacementToDictionary(replacementRules)
            defdictionary.update({'GroupName':groupname})
            definitionDictionaries.append(defdictionary)

        self.GaugeGroupDefinitions.append(definitionDictionaries)
        
        return definitionDictionaries
            

    def __init__(self):
        self.GaugeGroupDefinitions = []

class IndicesReader(Reader):

    def ExtractIndicesDefinitions(self):
        """
        Description:    This method takes as input all of the model file (given as an internal variable self.ModelFileText) and searches for the indices declarations (Range and Style).
                        
        Input:          The model fiel text stored as an internal variable self.ModelFileText
        Output:         The index declarations in the following form:
                        {'SU2W':{'IndexRange':'3','IndexStyle':'j'},'SU2D':{'IndexRange':'2','IndexStyle':'k'}}
        """
        
        #first take care of the index names and ranges
        #get the complete model file and strip off any spaces or line breaks
        fulldefinitionstring = self.ModelFileText
        fulldefinitionstring = self.RemoveSpaces(fulldefinitionstring)
        
        done = False

        indexnamesandranges = []
        definitionsDicts = {}

        while not done:
            Starti = fulldefinitionstring.find('IndexRange[Index[')
            if Starti == -1:
                done = True
            else:
                Namei = Starti + len('IndexRange[Index[')
                NameEndi = fulldefinitionstring[Namei:].find(']')
                RealNameEndi = NameEndi + Namei

                IndexName = fulldefinitionstring[Namei:RealNameEndi]
                
                Rangei = fulldefinitionstring[RealNameEndi:].find("Range[")
                RealRangei = RealNameEndi + Rangei + len("Range[")
                RangeEndi = fulldefinitionstring[RealRangei:].find("]")
                RealRangeEndi = RealRangei + RangeEndi

                IndexRange = fulldefinitionstring[RealRangei:RealRangeEndi]

                indexnamesandranges.append((IndexName,IndexRange))

                definitionsDicts.update({IndexName:{'IndexRange':IndexRange}})

                fulldefinitionstring = fulldefinitionstring[RealRangeEndi:]

        #now add the index styles

        fulldefinitionstring = self.ModelFileText
        fulldefinitionstring = self.RemoveSpaces(fulldefinitionstring)
        
        done = False

        while not done:
            Starti = fulldefinitionstring.find('IndexStyle[')
            if Starti == -1:
                done = True
            else:
                Namei = Starti + len('IndexStyle[')
                NameEndi = fulldefinitionstring[Namei:].find(',')
                RealNameEndi = NameEndi + Namei

                IndexName = fulldefinitionstring[Namei:RealNameEndi]
                
                RealStylei = RealNameEndi + 1
                StyleEndi = fulldefinitionstring[RealStylei:].find("]")
                RealStyleEndi = RealStylei + StyleEndi

                IndexStyle = fulldefinitionstring[RealStylei:RealStyleEndi]

                definitionsDicts[IndexName].update({'IndexStyle':IndexStyle})

                fulldefinitionstring = fulldefinitionstring[RealStyleEndi:]
        


        
        self.IndicesDefinitionsDict.update(definitionsDicts)
        return definitionsDicts


    def __init__(self):
        self.IndicesDefinitionsDict = {}

